@@ -7,6 +7,7 @@ module Network.Wai.Handler.Warp.HTTP2.HPACK (
   , hpackDecodeHeader
   , just
   , addNecessaryHeaders
+  , addHeader -- testing
   ) where
 
 import qualified Control.Exception as E
@@ -27,9 +28,12 @@ import Network.Wai.Handler.Warp.Types
 
 {-# INLINE addHeader #-}
 addHeader :: Token -> ByteString -> ValueTable -> TokenHeaderList -> TokenHeaderList
+addHeader t "" tbl ths = case getHeaderValue t tbl of
+    Nothing -> ths
+    _       -> filter ((/= tokenServer) . fst) ths
 addHeader t v tbl ths = case getHeaderValue t tbl of
     Nothing -> (t,v) : ths
-    Just _  -> ths
+    _       -> ths
 
 addNecessaryHeaders :: Context
                     -> Rspn
@@ -83,13 +87,16 @@ hpackDecodeHeader hdrblk Context{..} = do
 {-# INLINE checkRequestHeader #-}
 checkRequestHeader :: ValueTable -> Bool
 checkRequestHeader reqvt
+  | just mMethod (== "CONNECT") = mPath == Nothing && mScheme == Nothing
   | mStatus     /= Nothing      = False
   | mMethod     == Nothing      = False
+  | mScheme     == Nothing      = False
+  | mPath       == Nothing      = False
+  | mPath       == Just ""      = False
   | mAuthority  == Nothing      = False
   | mConnection /= Nothing      = False
   | just mTE (/= "trailers")    = False
-  | just mMethod (== "CONNECT") = mPath == Nothing && mScheme == Nothing
-  | otherwise                   = mPath /= Nothing
+  | otherwise                   = True
   where
     mStatus     = getHeaderValue tokenStatus reqvt
     mScheme     = getHeaderValue tokenScheme reqvt
