@@ -159,6 +159,8 @@ frameReceiver ctx mkreq recvN = loop 0 `E.catch` sendGoaway
                      when (ftyp == FrameHeaders) $ do
                          st <- readIORef $ streamState strm0
                          when (isHalfClosed st) $ E.throwIO $ ConnectionError StreamClosed "header must not be sent to half closed"
+                         -- Priority made an idele stream
+                         when (isIdle st) $ opened ctx strm0
                      return js
                  Nothing
                    | isResponse streamId -> return Nothing
@@ -166,19 +168,23 @@ frameReceiver ctx mkreq recvN = loop 0 `E.catch` sendGoaway
                          when (ftyp `notElem` [FrameHeaders,FramePriority]) $
                              E.throwIO $ ConnectionError ProtocolError "this frame is not allowed in an idel stream"
                          csid <- readIORef clientStreamId
-                         when (streamId <= csid) $
-                             E.throwIO $ ConnectionError ProtocolError "stream identifier must not decrease"
-                         when (ftyp == FrameHeaders) $ do
-                             writeIORef clientStreamId streamId
-                             cnt <- readIORef concurrency
-                             -- Checking the limitation of concurrency
-                             when (cnt >= maxConcurrency) $
-                                 E.throwIO $ StreamError RefusedStream streamId
-                         ws <- initialWindowSize <$> readIORef http2settings
-                         newstrm <- newStream streamId (fromIntegral ws)
-                         when (ftyp == FrameHeaders) $ opened ctx newstrm
-                         insert streamTable streamId newstrm
-                         return $ Just newstrm
+                         if streamId <= csid then do
+                             if ftyp == FramePriority then
+                                 return Nothing -- will be ignored
+                               else
+                                 E.throwIO $ ConnectionError ProtocolError "stream identifier must not decrease"
+                           else do
+                             when (ftyp == FrameHeaders) $ do
+                                 writeIORef clientStreamId streamId
+                                 cnt <- readIORef concurrency
+                                 -- Checking the limitation of concurrency
+                                 when (cnt >= maxConcurrency) $
+                                     E.throwIO $ StreamError RefusedStream streamId
+                             ws <- initialWindowSize <$> readIORef http2settings
+                             newstrm <- newStream streamId (fromIntegral ws)
+                             when (ftyp == FrameHeaders) $ opened ctx newstrm
+                             insert streamTable streamId newstrm
+                             return $ Just newstrm
 
     consume = void . recvN
 
@@ -191,14 +197,18 @@ initialFrame = settingsFrame id [(SettingsMaxConcurrentStreams,maxConcurrency)]
 ----------------------------------------------------------------
 
 control :: FrameTypeId -> FrameHeader -> ByteString -> Context -> IO Bool
-control FrameSettings header@FrameHeader{flags} bs Context{http2settings, controlQ,firstSettings} = do
+control FrameSettings header@FrameHeader{flags} bs Context{http2settings, controlQ, firstSettings, streamTable} = do
     SettingsFrame alist <- guardIt $ decodeSettingsFrame header bs
     case checkSettingsList alist of
         Just x  -> E.throwIO x
         Nothing -> return ()
     -- HTTP/2 Setting from a browser
     unless (testAck flags) $ do
+        oldws <- initialWindowSize <$> readIORef http2settings
         modifyIORef' http2settings $ \old -> updateSettings old alist
+        newws <- initialWindowSize <$> readIORef http2settings
+        let diff = newws - oldws
+        when (diff /= 0) $ updateAllStreamWindow (+ diff) streamTable
         let !frame = settingsFrame setAck []
         sent <- readIORef firstSettings
         let !setframe
@@ -210,7 +220,7 @@ control FrameSettings header@FrameHeader{flags} bs Context{http2settings, contro
 
 control FramePing FrameHeader{flags} bs Context{controlQ} =
     if testAck flags then
-        E.throwIO $ ConnectionError ProtocolError "the ack flag of this ping frame must not be set"
+        return True -- just ignore
       else do
         let !frame = pingFrame bs
         enqueueControl controlQ $ CFrame frame
