@@ -3,6 +3,7 @@
 {-# LANGUAGE TupleSections #-}
 {-# LANGUAGE RankNTypes #-}
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE BangPatterns #-}
 
 module Network.Wai.Handler.Warp.Response (
     sendResponse
@@ -10,6 +11,7 @@ module Network.Wai.Handler.Warp.Response (
   , warpVersion
   , hasBody
   , replaceHeader
+  , addServer -- testing
   ) where
 
 #ifndef MIN_VERSION_base
@@ -176,7 +178,8 @@ sendResponse settings conn ii req reqidxhdr src response = do
           | isHead                  -> RspNoBody
           | otherwise               -> RspStream fb needsChunked th
         ResponseRaw raw _           -> RspRaw raw src (T.tickle th)
-    ret = case response of
+    -- Make sure we don't hang on to 'response' (avoid space leak)
+    !ret = case response of
         ResponseFile    {} -> isPersist
         ResponseBuilder {} -> isKeepAlive
         ResponseStream  {} -> isKeepAlive
@@ -427,7 +430,11 @@ addDate getdate rspidxhdr hdrs = case rspidxhdr ! fromEnum ResDate of
 warpVersion :: String
 warpVersion = showVersion Paths_warp.version
 
+{-# INLINE addServer #-}
 addServer :: HeaderValue -> IndexedHeader -> H.ResponseHeaders -> H.ResponseHeaders
+addServer "" rspidxhdr hdrs = case rspidxhdr ! fromEnum ResServer of
+    Nothing -> hdrs
+    _       -> filter ((/= H.hServer) . fst) hdrs
 addServer serverName rspidxhdr hdrs = case rspidxhdr ! fromEnum ResServer of
     Nothing -> (H.hServer, serverName) : hdrs
     _       -> hdrs
